# Copyright (c) 2017–2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
IRC bot “chromebot”
"""

import asyncio, argparse, json, tempfile, time, random, os
from datetime import datetime
from urllib.parse import urlsplit
from enum import IntEnum, unique
from collections import defaultdict
from abc import abstractmethod
from functools import wraps
import bottom
import websockets

from .util import StrJsonEncoder

### helper functions ###
def prettyTimeDelta (seconds):
    """
    Pretty-print seconds to human readable string 1d 1h 1m 1s
    """
    seconds = int(seconds)
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    s = [(days, 'd'), (hours, 'h'), (minutes, 'm'), (seconds, 's')]
    s = filter (lambda x: x[0] != 0, s)
    return ' '.join (map (lambda x: '{}{}'.format (*x), s))

def prettyBytes (b):
    """
    Pretty-print bytes
    """
    prefixes = ['B', 'KiB', 'MiB', 'GiB', 'TiB']
    while b >= 1024 and len (prefixes) > 1:
        b /= 1024
        prefixes.pop (0)
    return f'{b:.1f} {prefixes[0]}'

def isValidUrl (s):
    url = urlsplit (s)
    if url.scheme and url.netloc and url.scheme in {'http', 'https'}:
        return s
    raise TypeError ()

class NonExitingArgumentParser (argparse.ArgumentParser):
    """ Argument parser that does not call exit(), suitable for interactive use """

    def exit (self, status=0, message=None):
        # should never be called
        pass

    def error (self, message):
        # if we use subparsers it’s important to return self, so we can show
        # the correct help
        raise Exception (self, message)

    def format_usage (self):
        return super().format_usage ().replace ('\n', ' ')

class Status(IntEnum):
    """ Job status """
    undefined = 0
    pending = 1
    running = 2
    aborted = 3
    finished = 4

# see https://arxiv.org/html/0901.4016 on how to build proquints (human
# pronouncable unique ids)
toConsonant = 'bdfghjklmnprstvz'
toVowel = 'aiou'

def u16ToQuint (v):
    """ Transform a 16 bit unsigned integer into a single quint """
    assert 0 <= v < 2**16
    # quints are “big-endian”
    return ''.join ([
            toConsonant[(v>>(4+2+4+2))&0xf],
            toVowel[(v>>(4+2+4))&0x3],
            toConsonant[(v>>(4+2))&0xf],
            toVowel[(v>>4)&0x3],
            toConsonant[(v>>0)&0xf],
            ])

def uintToQuint (v, length=2):
    """ Turn any integer into a proquint with fixed length """
    assert 0 <= v < 2**(length*16)

    return '-'.join (reversed ([u16ToQuint ((v>>(x*16))&0xffff) for x in range (length)]))

def makeJobId ():
    """ Create job id from time and randomness source """
    # allocate 48 bits for the time (in milliseconds) and add 16 random bits
    # at the end (just to be sure) for a total of 64 bits. Should be enough to
    # avoid collisions.
    randbits = 16
    stamp = (int (time.time ()*1000) << randbits) | random.randint (0, 2**randbits-1)
    return uintToQuint (stamp, 4)

class Job:
    """ Archival job """

    __slots__ = ('id', 'stats', 'rstats', 'started', 'finished', 'nick', 'status', 'process', 'url')

    def __init__ (self, url, nick):
        self.id = makeJobId ()
        self.stats = {}
        self.rstats = {}
        self.started = datetime.utcnow ()
        self.finished = None
        self.url = url
        # user who scheduled this job
        self.nick = nick
        self.status = Status.pending
        self.process = None

    def formatStatus (self):
        stats = self.stats
        rstats = self.rstats
        return (f"{self.url} ({self.id}) {self.status.name}. "
                f"{rstats.get ('have', 0)} pages finished, "
                f"{rstats.get ('pending', 0)} pending; "
                f"{stats.get ('crashed', 0)} crashed, "
                f"{stats.get ('requests', 0)} requests, "
                f"{stats.get ('failed', 0)} failed, "
                f"{prettyBytes (stats.get ('bytesRcv', 0))} received.")

@unique
class NickMode(IntEnum):
    # the actual numbers don’t matter, but their order must be strictly
    # increasing (with priviledge level)
    operator = 100
    voice = 10

    @classmethod
    def fromMode (cls, mode):
        return {'v': cls.voice, 'o': cls.operator}[mode]

    @classmethod
    def fromNickPrefix (cls, mode):
        return {'@': cls.operator, '+': cls.voice}[mode]

    @property
    def human (self):
        return {self.operator: 'operator', self.voice: 'voice'}[self]

class User:
    """ IRC user """
    __slots__ = ('name', 'modes')

    def __init__ (self, name, modes=None):
        self.name = name
        self.modes = modes or set ()

    def __eq__ (self, b):
        return self.name == b.name

    def __hash__ (self):
        return hash (self.name)

    def __repr__ (self):
        return f'<User {self.name} {self.modes}>'

    def hasPriv (self, p):
        if p is None:
            return True
        else:
            return self.modes and max (self.modes) >= p

    @classmethod
    def fromName (cls, name):
        """ Get mode and name from NAMES command """
        try:
            modes = {NickMode.fromNickPrefix (name[0])}
            name = name[1:]
        except KeyError:
            modes = set ()
        return cls (name, modes)

class ReplyContext:
    __slots__ = ('client', 'target', 'user')

    def __init__ (self, client, target, user):
        self.client = client
        self.target = target
        self.user = user

    def __call__ (self, message):
        self.client.send ('PRIVMSG', target=self.target,
                message=f'{self.user.name}: {message}')

class RefCountEvent:
    """
    Ref-counted event that triggers if a) armed and b) refcount drops to zero.
    
    Must be used as a context manager.
    """
    __slots__ = ('count', 'event', 'armed')

    def __init__ (self):
        self.armed = False
        self.count = 0
        self.event = asyncio.Event ()

    def __enter__ (self):
        self.count += 1
        self.event.clear ()

    def __exit__ (self, exc_type, exc_val, exc_tb):
        self.count -= 1
        if self.armed and self.count == 0:
            self.event.set ()

    async def wait (self):
        await self.event.wait ()

    def arm (self):
        self.armed = True
        if self.count == 0:
            self.event.set ()

class ArgparseBot (bottom.Client):
    """
    Simple IRC bot using argparse
    
    Tracks user’s modes, reconnects on disconnect
    """

    __slots__ = ('channels', 'nick', 'parser', 'users', '_quit')

    def __init__ (self, host, port, ssl, nick, logger, channels=None, loop=None):
        super().__init__ (host=host, port=port, ssl=ssl, loop=loop)
        self.channels = channels or []
        self.nick = nick
        # map channel -> nick -> user
        self.users = defaultdict (dict)
        self.logger = logger.bind (context=type (self).__name__)
        self.parser = self.getParser ()

        # bot does not accept new queries in shutdown mode, unless explicitly
        # permitted by the parser
        self._quit = RefCountEvent ()

        # register bottom event handler
        self.on('CLIENT_CONNECT', self.onConnect)
        self.on('PING', self.onKeepalive)
        self.on('PRIVMSG', self.onMessage)
        self.on('CLIENT_DISCONNECT', self.onDisconnect)
        self.on('RPL_NAMREPLY', self.onNameReply)
        self.on('CHANNELMODE', self.onMode)
        self.on('PART', self.onPart)
        self.on('JOIN', self.onJoin)
        # XXX: we would like to handle KICK, but bottom does not support that at the moment

    @abstractmethod
    def getParser (self):
        pass

    def cancel (self):
        self.logger.info ('cancel', uuid='1eb34aea-a854-4fec-90b2-7f8a3812a9cd')
        self._quit.arm ()
    
    async def run (self):
        await self.connect ()
        await self._quit.wait ()
        self.send ('QUIT', message='Bye.')
        await self.disconnect ()

    async def onConnect (self, **kwargs):
        self.logger.info ('connect', nick=self.nick, uuid='01f7b138-ea53-4609-88e9-61f3eca3e7e7')

        self.send('NICK', nick=self.nick)
        self.send('USER', user=self.nick, realname='https://github.com/PromyLOPh/crocoite')

        # Don't try to join channels until the server has
        # sent the MOTD, or signaled that there's no MOTD.
        done, pending = await asyncio.wait(
            [self.wait('RPL_ENDOFMOTD'), self.wait('ERR_NOMOTD')],
            loop=self.loop, return_when=asyncio.FIRST_COMPLETED)

        # Cancel whichever waiter's event didn't come in.
        for future in pending:
            future.cancel()

        for c in self.channels:
            self.logger.info ('join', channel=c, uuid='367063a5-9069-4025-907c-65ba88af8593')
            self.send ('JOIN', channel=c)
            # no need for NAMES here, server sends this automatically

    async def onNameReply (self, channel, users, **kwargs):
        # channels may be too big for a single message
        addusers = dict (map (lambda x: (x.name, x), map (User.fromName, users)))
        if channel not in self.users:
            self.users[channel] = addusers
        else:
            self.users[channel].update (addusers)

    @staticmethod
    def parseMode (mode):
        """ Parse mode strings like +a, -b, +a-b, -b+a, … """
        action = '+'
        ret = []
        for c in mode:
            if c in {'+', '-'}:
                action = c
            else:
                ret.append ((action, c))
        return ret

    async def onMode (self, channel, modes, params, **kwargs):
        if channel not in self.channels:
            return

        for (action, mode), nick in zip (self.parseMode (modes), params):
            try:
                m = NickMode.fromMode (mode)
                u = self.users[channel].get (nick, User (nick))
                if action == '+':
                    u.modes.add (m)
                elif action == '-':
                    u.modes.remove (m)
            except KeyError:
                # unknown mode, ignore
                pass

    async def onPart (self, nick, channel, **kwargs):
        if channel not in self.channels:
            return

        try:
            self.users[channel].pop (nick)
        except KeyError:
            # gone already
            pass

    async def onJoin (self, nick, channel, **kwargs):
        if channel not in self.channels:
            return

        self.users[channel][nick] = User (nick)

    async def onKeepalive (self, message, **kwargs):
        """ Ping received """
        self.send('PONG', message=message)

    async def onMessage (self, nick, target, message, **kwargs):
        """ Message received """
        if target in self.channels and message.startswith (self.nick + ':'):
            user = self.users[target].get (nick, User (nick))
            reply = ReplyContext (client=self, target=target, user=user)

            # channel message that starts with our nick
            command = message.split (' ')[1:]
            try:
                args = self.parser.parse_args (command)
            except Exception as e:
                reply (f'{e.args[1]} -- {e.args[0].format_usage ()}')
                return
            if not args or not hasattr (args, 'func'):
                reply (f'Sorry, I don’t understand {command}')
                return

            minPriv = getattr (args, 'minPriv', None)
            if self._quit.armed and not getattr (args, 'allowOnShutdown', False):
                reply ('Sorry, I’m shutting down and cannot accept your request right now.')
            elif not user.hasPriv (minPriv):
                reply (f'Sorry, you need the privilege {minPriv.human} to use this command.')
            else:
                with self._quit:
                    await args.func (user=user, args=args, reply=reply)

    async def onDisconnect (self, **kwargs):
        """ Auto-reconnect """
        self.logger.info ('disconnect', uuid='4c74b2c8-2403-4921-879d-2279ad85db72')
        while True:
            if not self._quit.armed:
                await asyncio.sleep (10, loop=self.loop)
                self.logger.info ('reconnect', uuid='c53555cb-e1a4-4b69-b1c9-3320269c19d7')
                try:
                    await self.connect ()
                finally:
                    break

def jobExists (func):
    """ Chromebot job exists """
    @wraps (func)
    async def inner (self, **kwargs):
        # XXX: not sure why it works with **kwargs, but not (user, args, reply)
        args = kwargs.get ('args')
        reply = kwargs.get ('reply')
        j = self.jobs.get (args.id, None)
        if not j:
            reply (f'Job {args.id} is unknown')
        else:
            ret = await func (self, job=j, **kwargs)
            return ret
    return inner

class Chromebot (ArgparseBot):
    __slots__ = ('jobs', 'tempdir', 'destdir', 'processLimit', 'blacklist', 'needVoice')

    def __init__ (self, host, port, ssl, nick, logger, channels=None,
            tempdir=None, destdir='.', processLimit=1,
            blacklist={}, needVoice=False, loop=None):
        self.needVoice = needVoice

        super().__init__ (host=host, port=port, ssl=ssl, nick=nick,
                logger=logger, channels=channels, loop=loop)

        self.jobs = {}
        self.tempdir = tempdir or tempfile.gettempdir()
        self.destdir = destdir
        self.processLimit = asyncio.Semaphore (processLimit)
        self.blacklist = blacklist

    def getParser (self):
        parser = NonExitingArgumentParser (prog=self.nick + ': ', add_help=False)
        subparsers = parser.add_subparsers(help='Sub-commands')

        archiveparser = subparsers.add_parser('a', help='Archive a site', add_help=False)
        #archiveparser.add_argument('--timeout', default=1*60*60, type=int, help='Maximum time for archival', metavar='SEC', choices=[60, 1*60*60, 2*60*60])
        #archiveparser.add_argument('--idle-timeout', default=10, type=int, help='Maximum idle seconds (i.e. no requests)', dest='idleTimeout', metavar='SEC', choices=[1, 10, 20, 30, 60])
        #archiveparser.add_argument('--max-body-size', default=None, type=int, dest='maxBodySize', help='Max body size', metavar='BYTES', choices=[1*1024*1024, 10*1024*1024, 100*1024*1024])
        archiveparser.add_argument('--concurrency', '-j', default=1, type=int, help='Parallel workers for this job', choices=range (1, 5))
        archiveparser.add_argument('--recursive', '-r', help='Enable recursion', choices=['0', '1', 'prefix'], default='0')
        archiveparser.add_argument('--insecure', '-k',
                help='Disable certificate checking', action='store_true')
        archiveparser.add_argument('url', help='Website URL', type=isValidUrl, metavar='URL')
        archiveparser.set_defaults (func=self.handleArchive,
                minPriv=NickMode.voice if self.needVoice else None)

        statusparser = subparsers.add_parser ('s', help='Get job status', add_help=False)
        statusparser.add_argument('id', help='Job id', metavar='UUID')
        statusparser.set_defaults (func=self.handleStatus, allowOnShutdown=True)

        abortparser = subparsers.add_parser ('r', help='Revoke/abort job', add_help=False)
        abortparser.add_argument('id', help='Job id', metavar='UUID')
        abortparser.set_defaults (func=self.handleAbort, allowOnShutdown=True,
                minPriv=NickMode.voice if self.needVoice else None)

        return parser

    def isBlacklisted (self, url):
        for k, v in self.blacklist.items():
            if k.match (url):
                return v
        return False

    async def handleArchive (self, user, args, reply):
        """ Handle the archive command """

        msg = self.isBlacklisted (args.url)
        if msg:
            reply (f'{args.url} cannot be queued: {msg}')
            return

        # make sure the job id is unique. Since ids are time-based we can just
        # wait.
        while True:
            j = Job (args.url, user.name)
            if j.id not in self.jobs:
                break
            time.sleep (0.01)
        self.jobs[j.id] = j

        logger = self.logger.bind (job=j.id)

        showargs = {
                'recursive': args.recursive,
                'concurrency': args.concurrency,
                }
        if args.insecure:
            showargs['insecure'] = args.insecure
        warcinfo = {'chromebot': {
                'jobid': j.id,
                'user': user.name,
                'queued': j.started,
                'url': args.url,
                'recursive': args.recursive,
                'concurrency': args.concurrency,
                }}
        grabCmd = ['crocoite-single']
        grabCmd.extend (['--warcinfo',
                '!' + json.dumps (warcinfo, cls=StrJsonEncoder)])
        if args.insecure:
            grabCmd.append ('--insecure')
        grabCmd.extend (['{url}', '{dest}'])
        # prefix warcinfo with !, so it won’t get expanded
        cmdline = ['crocoite',
                '--tempdir', self.tempdir,
                '--recursion', args.recursive,
                '--concurrency', str (args.concurrency),
                args.url,
                os.path.join (self.destdir,
                        j.id + '-{host}-{date}-{seqnum}.warc.gz'),
                '--'] + grabCmd

        strargs = ', '.join (map (lambda x: '{}={}'.format (*x), showargs.items ()))
        reply (f'{args.url} has been queued as {j.id} with {strargs}')
        logger.info ('queue', user=user.name, url=args.url, cmdline=cmdline,
                uuid='36cc34a6-061b-4cc5-84a9-4ab6552c8d75')

        async with self.processLimit:
            if j.status == Status.pending:
                # job was not aborted
                j.process = await asyncio.create_subprocess_exec (*cmdline,
                        stdout=asyncio.subprocess.PIPE,
                        stderr=asyncio.subprocess.DEVNULL,
                        stdin=asyncio.subprocess.DEVNULL,
                        start_new_session=True, limit=100*1024*1024)
                while True:
                    data = await j.process.stdout.readline ()
                    if not data:
                        break

                    # job is marked running after the first message is received from it
                    if j.status == Status.pending:
                        logger.info ('start', uuid='46e62d60-f498-4ab0-90e1-d08a073b10fb')
                        j.status = Status.running

                    data = json.loads (data)
                    msgid = data.get ('uuid')
                    if msgid == '24d92d16-770e-4088-b769-4020e127a7ff':
                        j.stats = data
                    elif msgid == '5b8498e4-868d-413c-a67e-004516b8452c':
                        j.rstats = data

                    # forward message, so the dashboard can use it
                    logger.info ('message',
                            uuid='5c0f9a11-dcd8-4182-a60f-54f4d3ab3687',
                            data=data)
                code = await j.process.wait ()

        if j.status == Status.running:
            logger.info ('finish', uuid='7b40ffbb-faab-4224-90ed-cd4febd8f7ec')
            j.status = Status.finished
        j.finished = datetime.utcnow ()

        stats = j.stats
        rstats = j.rstats
        reply (j.formatStatus ())

    @jobExists
    async def handleStatus (self, user, args, reply, job):
        """ Handle status command """

        rstats = job.rstats
        reply (job.formatStatus ())

    @jobExists
    async def handleAbort (self, user, args, reply, job):
        """ Handle abort command """

        if job.status not in {Status.pending, Status.running}:
            reply ('This job is not running.')
            return

        job.status = Status.aborted
        self.logger.info ('abort', job=job.id, user=user.name,
                uuid='865b3b3e-a54a-4a56-a545-f38a37bac295')
        if job.process and job.process.returncode is None:
            job.process.terminate ()

class Dashboard:
    __slots__ = ('fd', 'clients', 'loop', 'log', 'maxLog', 'pingInterval', 'pingTimeout')
    # these messages will not be forwarded to the browser
    ignoreMsgid = {
            # connect
            '01f7b138-ea53-4609-88e9-61f3eca3e7e7',
            # join
            '367063a5-9069-4025-907c-65ba88af8593',
            # disconnect
            '4c74b2c8-2403-4921-879d-2279ad85db72',
            # reconnect
            'c53555cb-e1a4-4b69-b1c9-3320269c19d7',
            }

    def __init__ (self, fd, loop, maxLog=1000, pingInterval=30, pingTimeout=10):
        self.fd = fd
        self.clients = set ()
        self.loop = loop
        # log buffer
        self.log = []
        self.maxLog = maxLog
        self.pingInterval = pingInterval
        self.pingTimeout = pingTimeout

    async def client (self, websocket, path):
        self.clients.add (websocket)
        try:
            for l in self.log:
                buf = json.dumps (l)
                await websocket.send (buf)
            while True:
                try:
                    msg = await asyncio.wait_for (websocket.recv(), timeout=self.pingInterval)
                except asyncio.TimeoutError:
                    try:
                        pong = await websocket.ping()
                        await asyncio.wait_for (pong, timeout=self.pingTimeout)
                    except asyncio.TimeoutError:
                        break
                except websockets.exceptions.ConnectionClosed:
                    break
        finally:
            self.clients.remove (websocket)

    def handleStdin (self):
        buf = self.fd.readline ()
        if not buf:
            return

        try:
            data = json.loads (buf)
        except json.decoder.JSONDecodeError:
            # ignore invalid
            return
        msgid = data['uuid']

        if msgid in self.ignoreMsgid:
            return

        # a few messages may contain sensitive information that we want to hide
        if msgid == '36cc34a6-061b-4cc5-84a9-4ab6552c8d75':
            # queue
            del data['cmdline']
        elif msgid == '5c0f9a11-dcd8-4182-a60f-54f4d3ab3687':
            nesteddata = data['data']
            nestedmsgid = nesteddata['uuid']
            if nestedmsgid == 'd1288fbe-8bae-42c8-af8c-f2fa8b41794f':
                del nesteddata['command']
            
        buf = json.dumps (data)
        for c in self.clients:
            # XXX can’t await here
            asyncio.ensure_future (c.send (buf))

        self.log.append (data)
        while len (self.log) > self.maxLog:
            self.log.pop (0)

    def run (self, host='localhost', port=6789):
        self.loop.add_reader (self.fd, self.handleStdin)
        return websockets.serve(self.client, host, port)




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import asyncio, os, yaml, re, math, struct
from functools import partial
from operator import attrgetter

import pytest
from yarl import URL
from aiohttp import web

import pkg_resources
from .logger import Logger
from .devtools import Process
from .behavior import Scroll, Behavior, ExtractLinks, ExtractLinksEvent, Crash, \
        Screenshot, ScreenshotEvent, DomSnapshot, DomSnapshotEvent, mapOrIgnore
from .controller import SinglePageController, EventHandler
from .devtools import Crashed

with pkg_resources.resource_stream (__name__, os.path.join ('data', 'click.yaml')) as fd:
    sites = list (yaml.safe_load_all (fd))
clickParam = []
for o in sites:
    for s in o['selector']:
        for u in s.get ('urls', []):
            clickParam.append ((u, s['selector']))

class ClickTester (Behavior):
    """
    Test adapter checking a given selector exists after loading the page
    """

    __slots__ = ('selector', )

    name = 'testclick'

    def __init__ (self, loader, logger, selector):
        super ().__init__ (loader, logger)
        self.selector = selector

    async def onfinish (self):
        tab = self.loader.tab
        results = await tab.DOM.getDocument ()
        rootNode = results['root']['nodeId']
        results = await tab.DOM.querySelectorAll (nodeId=rootNode, selector=self.selector)
        assert results['nodeIds'], self.selector

        # XXX: this is not true for every element we click. Github uses <button
        # type=submit> and <form> without an event listener on the <button>
#        # verify that an event listener exists
#        for nid in results['nodeIds']:
#            obj = (await tab.DOM.resolveNode (nodeId=nid))['object']
#            assert obj['type'] == 'object'
#            listeners = (await tab.DOMDebugger.getEventListeners (objectId=obj['objectId']))['listeners']
#            assert any (map (lambda x: x['type'] == 'click', listeners)), listeners

        return
        yield # pragma: no cover

@pytest.mark.parametrize("url,selector", clickParam)
@pytest.mark.asyncio
@pytest.mark.xfail(reason='depends on network access')
async def test_click_selectors (url, selector):
    """
    Make sure the CSS selector exists on an example url
    """
    logger = Logger ()
    # Some selectors are loaded dynamically and require scrolling
    controller = SinglePageController (url=url, logger=logger,
            service=Process (),
            behavior=[Scroll, partial(ClickTester, selector=selector)])
    await controller.run ()

matchParam = []
for o in sites:
    for s in o['selector']:
        for u in s.get ('urls', []):
            matchParam.append ((o['match'], URL (u)))

@pytest.mark.parametrize("match,url", matchParam)
@pytest.mark.asyncio
async def test_click_match (match, url):
    """ Test urls must match """
    # keep this aligned with click.js
    assert re.match (match, url.host, re.I)


class AccumHandler (EventHandler):
    """ Test adapter that accumulates all incoming items """
    __slots__ = ('data')

    def __init__ (self):
        super().__init__ ()
        self.data = []

    async def push (self, item):
        self.data.append (item)

async def simpleServer (url, response):
    async def f (req):
        return web.Response (body=response, status=200, content_type='text/html', charset='utf-8')

    app = web.Application ()
    app.router.add_route ('GET', url.path, f)
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, url.host, url.port)
    await site.start()
    return runner

@pytest.mark.asyncio
async def test_extract_links ():
    """
    Make sure the CSS selector exists on an example url
    """

    url = URL.build (scheme='http', host='localhost', port=8080)
    runner = await simpleServer (url, """<html><head></head>
            <body>
            <div>
                <a href="/relative">foo</a>
                <a href="http://example.com/absolute/">foo</a>
                <a href="https://example.com/absolute/secure">foo</a>
                <a href="#anchor">foo</a>
                <a href="http://neue_preise_f%c3%bcr_zahnimplantate_k%c3%b6nnten_sie_%c3%bcberraschen">foo</a>

                <a href="/hidden/visibility" style="visibility: hidden">foo</a>
                <a href="/hidden/display" style="display: none">foo</a>
                <div style="display: none">
                <a href="/hidden/display/insidediv">foo</a>
                </div>
                <!--<a href="/hidden/comment">foo</a>-->

                <p><img src="shapes.png" usemap="#shapes">
                 <map name="shapes"><area shape=rect coords="50,50,100,100" href="/map/rect"></map></p>
            </div>
            </body></html>""")

    try:
        handler = AccumHandler ()
        logger = Logger ()
        controller = SinglePageController (url=url, logger=logger,
                service=Process (), behavior=[ExtractLinks], handler=[handler])
        await controller.run ()

        links = []
        for d in handler.data:
            if isinstance (d, ExtractLinksEvent):
                links.extend (d.links)
        assert sorted (links) == sorted ([
                url.with_path ('/relative'),
                url.with_fragment ('anchor'),
                URL ('http://example.com/absolute/'),
                URL ('https://example.com/absolute/secure'),
                url.with_path ('/hidden/visibility'), # XXX: shall we ignore these as well?
                url.with_path ('/map/rect'),
                ])
    finally:
        await runner.cleanup ()

@pytest.mark.asyncio
async def test_crash ():
    """
    Crashing through Behavior works?
    """

    url = URL.build (scheme='http', host='localhost', port=8080)
    runner = await simpleServer (url, '<html></html>')

    try:
        logger = Logger ()
        controller = SinglePageController (url=url, logger=logger,
                service=Process (), behavior=[Crash])
        with pytest.raises (Crashed):
            await controller.run ()
    finally:
        await runner.cleanup ()

@pytest.mark.asyncio
async def test_screenshot ():
    """
    Make sure screenshots are taken and have the correct dimensions. We can’t
    and don’t want to check their content.
    """
    # ceil(0) == 0, so starting with 1
    for expectHeight in (1, Screenshot.maxDim, Screenshot.maxDim+1, Screenshot.maxDim*2+Screenshot.maxDim//2):
        url = URL.build (scheme='http', host='localhost', port=8080)
        runner = await simpleServer (url, f'<html><body style="margin: 0; padding: 0;"><div style="height: {expectHeight}"></div></body></html>')

        try:
            handler = AccumHandler ()
            logger = Logger ()
            controller = SinglePageController (url=url, logger=logger,
                    service=Process (), behavior=[Screenshot], handler=[handler])
            await controller.run ()

            screenshots = list (filter (lambda x: isinstance (x, ScreenshotEvent), handler.data))
            assert len (screenshots) == math.ceil (expectHeight/Screenshot.maxDim)
            totalHeight = 0
            for s in screenshots:
                assert s.url == url
                # PNG ident is fixed, IHDR is always the first chunk
                assert s.data.startswith (b'\x89PNG\r\n\x1a\n\x00\x00\x00\x0dIHDR')
                width, height = struct.unpack ('>II', s.data[16:24])
                assert height <= Screenshot.maxDim
                totalHeight += height
            # screenshot height is at least canvas height (XXX: get hardcoded
            # value from devtools.Process)
            assert totalHeight == max (expectHeight, 1080)
        finally:
            await runner.cleanup ()

@pytest.mark.asyncio
async def test_dom_snapshot ():
    """
    Behavior plug-in works, <canvas> is replaced by static image, <script> is
    stripped. Actual conversion from Chrome DOM to HTML is validated by module
    .test_html
    """

    url = URL.build (scheme='http', host='localhost', port=8080)
    runner = await simpleServer (url, f'<html><body><p>ÄÖÜäöü</p><script>alert("yes");</script><canvas id="canvas" width="1" height="1">Alternate text.</canvas></body></html>')

    try:
        handler = AccumHandler ()
        logger = Logger ()
        controller = SinglePageController (url=url, logger=logger,
                service=Process (), behavior=[DomSnapshot], handler=[handler])
        await controller.run ()

        snapshots = list (filter (lambda x: isinstance (x, DomSnapshotEvent), handler.data))
        assert len (snapshots) == 1
        doc = snapshots[0].document
        assert doc.startswith ('<HTML><HEAD><meta charset=utf-8></HEAD><BODY><P>ÄÖÜäöü</P><IMG id=canvas width=1 height=1 src="data:image/png;base64,'.encode ('utf-8'))
        assert doc.endswith ('></BODY></HTML>'.encode ('utf-8'))
    finally:
        await runner.cleanup ()

def test_mapOrIgnore ():
    def fail (x):
        if x < 50:
            raise Exception ()
        return x+1

    assert list (mapOrIgnore (fail, range (100))) == list (range (51, 101))




from setuptools import setup

setup(
    name='crocoite',
    version='1.0.0',
    author='Lars-Dominik Braun',
    author_email='lars+crocoite@6xq.net',
    url='https://6xq.net/crocoite/',
    packages=['crocoite'],
    license='LICENSE.txt',
    description='Save website to WARC using Google Chrome.',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    install_requires=[
        'warcio',
        'html5lib>=0.999999999',
        'bottom',
        'pytz',
        'websockets',
        'aiohttp',
        'PyYAML',
        'yarl',
        'multidict',
    ],
    extras_require={
        'manhole': ['manhole>=1.6'],
    },
    entry_points={
    'console_scripts': [
            # the main executable
            'crocoite = crocoite.cli:recursive',
            # backend helper
            'crocoite-single = crocoite.cli:single',
            # irc bot and dashboard
            'crocoite-irc = crocoite.cli:irc',
            'crocoite-irc-dashboard = crocoite.cli:dashboard',
            # misc tools
            'crocoite-merge-warc = crocoite.tools:mergeWarcCli',
            'crocoite-extract-screenshot = crocoite.tools:extractScreenshot',
            'crocoite-errata = crocoite.tools:errata',
            ],
    },
    package_data={
            'crocoite': ['data/*'],
    },
    setup_requires=['pytest-runner'],
    tests_require=["pytest", 'pytest-asyncio', 'pytest-cov', 'hypothesis'],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



# Copyright (c) 2017–2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Controller classes, handling actions required for archival
"""

import time, tempfile, asyncio, json, os, shutil
from itertools import islice
from datetime import datetime
from operator import attrgetter
from abc import ABC, abstractmethod
from yarl import URL

from . import behavior as cbehavior
from .browser import SiteLoader, RequestResponsePair, PageIdle, FrameNavigated
from .util import getFormattedViewportMetrics, getSoftwareInfo
from .behavior import ExtractLinksEvent

class ControllerSettings:
    __slots__ = ('idleTimeout', 'timeout', 'insecure')

    def __init__ (self, idleTimeout=2, timeout=10, insecure=False):
        self.idleTimeout = idleTimeout
        self.timeout = timeout
        self.insecure = insecure

    def toDict (self):
        return dict (
                idleTimeout=self.idleTimeout,
                timeout=self.timeout,
                insecure=self.insecure,
                )

defaultSettings = ControllerSettings ()

class EventHandler (ABC):
    """ Abstract base class for event handler """

    __slots__ = ()

    @abstractmethod
    async def push (self, item):
        raise NotImplementedError ()

class StatsHandler (EventHandler):
    __slots__ = ('stats', )

    def __init__ (self):
        self.stats = {'requests': 0, 'finished': 0, 'failed': 0, 'bytesRcv': 0}

    async def push (self, item):
        if isinstance (item, RequestResponsePair):
            self.stats['requests'] += 1
            if not item.response:
                self.stats['failed'] += 1
            else:
                self.stats['finished'] += 1
                self.stats['bytesRcv'] += item.response.bytesReceived

class LogHandler (EventHandler):
    """ Handle items by logging information about them """

    __slots__ = ('logger', )

    def __init__ (self, logger):
        self.logger = logger.bind (context=type (self).__name__)

    async def push (self, item):
        if isinstance (item, ExtractLinksEvent):
            # limit number of links per message, so json blob won’t get too big
            it = iter (item.links)
            limit = 100
            while True:
                limitlinks = list (islice (it, 0, limit))
                if not limitlinks:
                    break
                self.logger.info ('extracted links', context=type (item).__name__,
                        uuid='8ee5e9c9-1130-4c5c-88ff-718508546e0c', links=limitlinks)


class ControllerStart:
    __slots__ = ('payload', )

    def __init__ (self, payload):
        self.payload = payload

class IdleStateTracker (EventHandler):
    """ Track SiteLoader’s idle state by listening to PageIdle events """

    __slots__ = ('_idle', '_loop', '_idleSince')

    def __init__ (self, loop):
        self._idle = True
        self._loop = loop

        self._idleSince = self._loop.time ()

    async def push (self, item):
        if isinstance (item, PageIdle):
            self._idle = bool (item)
            if self._idle:
                self._idleSince = self._loop.time ()

    async def wait (self, timeout):
        """ Wait until page has been idle for at least timeout seconds. If the
        page has been idle before calling this function it may return
        immediately. """

        assert timeout > 0
        while True:
            if self._idle:
                now = self._loop.time ()
                sleep = timeout-(now-self._idleSince)
                if sleep <= 0:
                    break
            else:
                # not idle, check again after timeout expires
                sleep = timeout
            await asyncio.sleep (sleep)

class InjectBehaviorOnload (EventHandler):
    """ Control behavior script injection based on frame navigation messages.
    When a page is reloaded (for whatever reason), the scripts need to be
    reinjected. """

    __slots__ = ('controller', '_loaded')

    def __init__ (self, controller):
        self.controller = controller
        self._loaded = False

    async def push (self, item):
        if isinstance (item, FrameNavigated):
            await self._runon ('load')
            self._loaded = True

    async def stop (self):
        if self._loaded:
            await self._runon ('stop')

    async def finish (self):
        if self._loaded:
            await self._runon ('finish')

    async def _runon (self, method):
        controller = self.controller
        for b in controller._enabledBehavior:
            f = getattr (b, 'on' + method)
            async for item in f ():
                await controller.processItem (item)

class SinglePageController:
    """
    Archive a single page url.

    Dispatches between producer (site loader and behavior scripts) and consumer
    (stats, warc writer).
    """

    __slots__ = ('url', 'service', 'behavior', 'settings', 'logger', 'handler',
            'warcinfo', '_enabledBehavior')

    def __init__ (self, url, logger, \
            service, behavior=cbehavior.available, \
            settings=defaultSettings, handler=None, \
            warcinfo=None):
        self.url = url
        self.service = service
        self.behavior = behavior
        self.settings = settings
        self.logger = logger.bind (context=type (self).__name__, url=url)
        self.handler = handler or []
        self.warcinfo = warcinfo

    async def processItem (self, item):
        for h in self.handler:
            await h.push (item)

    async def run (self):
        logger = self.logger
        async def processQueue ():
            async for item in l:
                await self.processItem (item)

        idle = IdleStateTracker (asyncio.get_event_loop ())
        self.handler.append (idle)
        behavior = InjectBehaviorOnload (self)
        self.handler.append (behavior)

        async with self.service as browser, SiteLoader (browser, logger=logger) as l:
            handle = asyncio.ensure_future (processQueue ())
            timeoutProc = asyncio.ensure_future (asyncio.sleep (self.settings.timeout))

            # configure browser
            tab = l.tab
            await tab.Security.setIgnoreCertificateErrors (ignore=self.settings.insecure)

            # not all behavior scripts are allowed for every URL, filter them
            self._enabledBehavior = list (filter (lambda x: self.url in x,
                    map (lambda x: x (l, logger), self.behavior)))

            version = await tab.Browser.getVersion ()
            payload = {
                    'software': getSoftwareInfo (),
                    'browser': {
                        'product': version['product'],
                        'useragent': version['userAgent'],
                        'viewport': await getFormattedViewportMetrics (tab),
                        },
                    'tool': 'crocoite-single', # not the name of the cli utility
                    'parameters': {
                        'url': self.url,
                        'idleTimeout': self.settings.idleTimeout,
                        'timeout': self.settings.timeout,
                        'behavior': list (map (attrgetter('name'), self._enabledBehavior)),
                        'insecure': self.settings.insecure,
                        },
                    }
            if self.warcinfo:
                payload['extra'] = self.warcinfo
            await self.processItem (ControllerStart (payload))

            await l.navigate (self.url)

            idleProc = asyncio.ensure_future (idle.wait (self.settings.idleTimeout))
            while True:
                try:
                    finished, pending = await asyncio.wait([idleProc, timeoutProc, handle],
                            return_when=asyncio.FIRST_COMPLETED)
                except asyncio.CancelledError:
                    idleProc.cancel ()
                    timeoutProc.cancel ()
                    break

                if handle in finished:
                    # something went wrong while processing the data
                    logger.error ('fetch failed',
                        uuid='43a0686a-a3a9-4214-9acd-43f6976f8ff3')
                    idleProc.cancel ()
                    timeoutProc.cancel ()
                    handle.result ()
                    assert False # previous line should always raise Exception
                elif timeoutProc in finished:
                    # global timeout
                    logger.debug ('global timeout',
                            uuid='2f858adc-9448-4ace-94b4-7cd1484c0728')
                    idleProc.cancel ()
                    timeoutProc.result ()
                    break
                elif idleProc in finished:
                    # idle timeout
                    logger.debug ('idle timeout',
                            uuid='90702590-94c4-44ef-9b37-02a16de444c3')
                    idleProc.result ()
                    timeoutProc.cancel ()
                    break

            await behavior.stop ()
            await tab.Page.stopLoading ()
            await asyncio.sleep (1)
            await behavior.finish ()

            # wait until loads from behavior scripts are done and browser is
            # idle for at least 1 second
            try:
                await asyncio.wait_for (idle.wait (1), timeout=1)
            except (asyncio.TimeoutError, asyncio.CancelledError):
                pass

            if handle.done ():
                handle.result ()
            else:
                handle.cancel ()

class SetEntry:
    """ A object, to be used with sets, that compares equality only on its
    primary property. """
    def __init__ (self, value, **props):
        self.value = value
        for k, v in props.items ():
            setattr (self, k, v)

    def __eq__ (self, b):
        assert isinstance (b, SetEntry)
        return self.value == b.value

    def __hash__ (self):
        return hash (self.value)

    def __repr__ (self):
        return f'<SetEntry {self.value!r}>'

class RecursionPolicy:
    """ Abstract recursion policy """

    __slots__ = ()

    def __call__ (self, urls):
        raise NotImplementedError

class DepthLimit (RecursionPolicy):
    """
    Limit recursion by depth.
    
    depth==0 means no recursion, depth==1 is the page and outgoing links
    """

    __slots__ = ('maxdepth', )

    def __init__ (self, maxdepth=0):
        self.maxdepth = maxdepth

    def __call__ (self, urls):
        newurls = set ()
        for u in urls:
            if u.depth <= self.maxdepth:
                newurls.add (u)
        return newurls

    def __repr__ (self):
        return f'<DepthLimit {self.maxdepth}>'

class PrefixLimit (RecursionPolicy):
    """
    Limit recursion by prefix
    
    i.e. prefix=http://example.com/foo
    ignored: http://example.com/bar http://offsite.example/foo
    accepted: http://example.com/foobar http://example.com/foo/bar
    """

    __slots__ = ('prefix', )

    def __init__ (self, prefix):
        self.prefix = prefix

    def __call__ (self, urls):
        return set (filter (lambda u: str(u.value).startswith (str (self.prefix)), urls))

def hasTemplate (s):
    """ Return True if string s has string templates """
    return '{' in s and '}' in s

class RecursiveController:
    """
    Simple recursive controller

    Visits links acording to policy
    """

    __slots__ = ('url', 'output', 'command', 'logger', 'policy', 'have',
            'pending', 'stats', 'tempdir', 'running', 'concurrency',
            'copyLock')

    SCHEME_WHITELIST = {'http', 'https'}

    def __init__ (self, url, output, command, logger,
            tempdir=None, policy=DepthLimit (0), concurrency=1):
        self.url = url
        self.output = output
        self.command = command
        self.logger = logger.bind (context=type(self).__name__, seedurl=url)
        self.policy = policy
        self.tempdir = tempdir
        # A lock if only a single output file (no template) is requested
        self.copyLock = None if hasTemplate (output) else asyncio.Lock ()
        # some sanity checks. XXX move to argparse?
        if self.copyLock and os.path.exists (self.output):
                raise ValueError ('Output file exists')
        # tasks currently running
        self.running = set ()
        # max number of tasks running
        self.concurrency = concurrency
        # keep in sync with StatsHandler
        self.stats = {'requests': 0, 'finished': 0, 'failed': 0, 'bytesRcv': 0, 'crashed': 0, 'ignored': 0}

    async def fetch (self, entry, seqnum):
        """
        Fetch a single URL using an external command

        command is usually crocoite-single
        """

        assert isinstance (entry, SetEntry)

        url = entry.value
        depth = entry.depth
        logger = self.logger.bind (url=url)

        def formatCommand (e):
            # provide means to disable variable expansion
            if e.startswith ('!'):
                return e[1:]
            else:
                return e.format (url=url, dest=dest.name)

        def formatOutput (p):
            return p.format (host=url.host,
                    date=datetime.utcnow ().isoformat (), seqnum=seqnum)

        def logStats ():
            logger.info ('stats', uuid='24d92d16-770e-4088-b769-4020e127a7ff', **self.stats)

        if url.scheme not in self.SCHEME_WHITELIST:
            self.stats['ignored'] += 1
            logStats ()
            self.logger.warning ('scheme not whitelisted', url=url,
                    uuid='57e838de-4494-4316-ae98-cd3a2ebf541b')
            return

        dest = tempfile.NamedTemporaryFile (dir=self.tempdir,
                prefix=__package__, suffix='.warc.gz', delete=False)
        command = list (map (formatCommand, self.command))
        logger.info ('fetch', uuid='d1288fbe-8bae-42c8-af8c-f2fa8b41794f',
                command=command)
        try:
            process = await asyncio.create_subprocess_exec (*command,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.DEVNULL,
                    stdin=asyncio.subprocess.DEVNULL,
                    start_new_session=True, limit=100*1024*1024)
            while True:
                data = await process.stdout.readline ()
                if not data:
                    break
                data = json.loads (data)
                uuid = data.get ('uuid')
                if uuid == '8ee5e9c9-1130-4c5c-88ff-718508546e0c':
                    links = set (self.policy (map (lambda x: SetEntry (URL(x).with_fragment(None), depth=depth+1), data.get ('links', []))))
                    links.difference_update (self.have)
                    self.pending.update (links)
                elif uuid == '24d92d16-770e-4088-b769-4020e127a7ff':
                    for k in self.stats.keys ():
                        self.stats[k] += data.get (k, 0)
                    logStats ()
        except asyncio.CancelledError:
            # graceful cancellation
            process.send_signal (signal.SIGINT)
        except Exception as e:
            process.kill ()
            raise e
        finally:
            code = await process.wait()
            if code == 0:
                if self.copyLock is None:
                    # atomically move once finished
                    lastDestpath = None
                    while True:
                        # XXX: must generate a new name every time, otherwise
                        # this loop never terminates
                        destpath = formatOutput (self.output)
                        assert destpath != lastDestpath
                        lastDestpath = destpath

                        # python does not have rename(…, …, RENAME_NOREPLACE),
                        # but this is safe nontheless, since we’re
                        # single-threaded
                        if not os.path.exists (destpath):
                            # create the directory, so templates like
                            # /{host}/{date}/… are possible
                            os.makedirs (os.path.dirname (destpath), exist_ok=True)
                            os.rename (dest.name, destpath)
                            break
                else:
                    # atomically (in the context of this process) append to
                    # existing file
                    async with self.copyLock:
                        with open (dest.name, 'rb') as infd, \
                                open (self.output, 'ab') as outfd:
                            shutil.copyfileobj (infd, outfd)
                        os.unlink (dest.name)
            else:
                self.stats['crashed'] += 1
                logStats ()

    async def run (self):
        def log ():
            # self.have includes running jobs
            self.logger.info ('recursing',
                    uuid='5b8498e4-868d-413c-a67e-004516b8452c',
                    pending=len (self.pending),
                    have=len (self.have)-len(self.running),
                    running=len (self.running))

        seqnum = 1
        try:
            self.have = set ()
            self.pending = set ([SetEntry (self.url, depth=0)])

            while self.pending:
                # since pending is a set this picks a random item, which is fine
                u = self.pending.pop ()
                self.have.add (u)
                t = asyncio.ensure_future (self.fetch (u, seqnum))
                self.running.add (t)
                seqnum += 1

                log ()

                if len (self.running) >= self.concurrency or not self.pending:
                    done, pending = await asyncio.wait (self.running,
                            return_when=asyncio.FIRST_COMPLETED)
                    self.running.difference_update (done)
                    # propagate exceptions
                    for r in done:
                        r.result ()
        except asyncio.CancelledError:
            self.logger.info ('cancel',
                    uuid='d58154c8-ec27-40f2-ab9e-e25c1b21cd88',
                    pending=len (self.pending),
                    have=len (self.have)-len (self.running),
                    running=len (self.running))
        finally:
            done = await asyncio.gather (*self.running,
                    return_exceptions=True)
            # propagate exceptions
            for r in done:
                if isinstance (r, Exception):
                    raise r
            self.running = set ()
            log ()




import pytest
from .logger import Logger, Consumer, NullConsumer, Level, DatetimeConsumer

@pytest.fixture
def logger ():
    return Logger (consumer=[NullConsumer (), DatetimeConsumer ()])

class QueueConsumer (Consumer):
    def __init__ (self):
        self.data = []

    def __call__ (self, **kwargs):
        self.data.append (kwargs)
        return kwargs

def test_bind (logger):
    # simple bind
    logger = logger.bind (foo='bar')
    ret = logger.debug ()
    assert ret['foo'] == 'bar'

    # additional
    ret = logger.debug (bar='baz')
    assert ret['foo'] == 'bar'
    assert ret['bar'] == 'baz'
    
    # override
    ret = logger.debug (foo='baz')
    assert ret['foo'] == 'baz'

    # unbind
    logger = logger.unbind (foo=None)
    ret = logger.debug ()
    assert 'foo' not in ret

def test_consumer (logger):
    c = QueueConsumer ()
    logger.connect (c)
    ret = logger.debug (foo='bar')
    assert len (c.data) == 1
    assert c.data[0] == ret
    assert ret['foo'] == 'bar'
    c.data = []

    # inheritance
    logger = logger.bind (inherit=1)
    ret = logger.debug (foo='bar')
    assert len (c.data) == 1
    assert c.data[0] == ret
    assert ret['foo'] == 'bar'
    assert ret['inherit'] == 1
    c.data = []

    # removal
    logger.disconnect (c)
    ret = logger.debug (foo='bar')
    assert len (c.data) == 0
    assert ret['foo'] == 'bar'
    assert ret['inherit'] == 1

def test_multiarg (logger):
    # single argument
    ret = logger.debug('maybe', foo='bar')
    assert ret['msg'] == 'maybe'
    assert ret['foo'] == 'bar'

    # multi arguments
    ret = logger.debug('may', 'be', foo='bar')
    assert ret['msg'] == ('may', 'be')
    assert ret['foo'] == 'bar'

def test_call (logger):
    for level in ('debug', Level.DEBUG):
        ret = logger(level, 'arg1', 'arg2', foo='bar')
        assert ret['level'] == Level.DEBUG
        assert ret['msg'] == ('arg1', 'arg2')
        assert ret['foo'] == 'bar'

def test_datetime (logger):
    ret = logger.debug()
    assert 'date' in ret

def test_independence ():
    """ Make sure two instances are completely independent """
    l1 = Logger ()
    c = QueueConsumer ()
    l1.connect (c)
    l2 = Logger ()
    l2.info (nothing='nothing')
    assert not c.data




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Random utility functions
"""

import random, sys, platform, os, json, urllib
from datetime import datetime
import hashlib, pkg_resources

from yarl import URL

class StrJsonEncoder (json.JSONEncoder):
    """ JSON encoder that turns unknown classes into a string and thus never
    fails """
    def default (self, obj):
        if isinstance (obj, datetime):
            return obj.isoformat ()

        # make sure serialization always succeeds
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            return str (obj)

async def getFormattedViewportMetrics (tab):
    layoutMetrics = await tab.Page.getLayoutMetrics ()
    # XXX: I’m not entirely sure which one we should use here
    viewport = layoutMetrics['layoutViewport']
    return f"{viewport['clientWidth']}x{viewport['clientHeight']}"

def getSoftwareInfo ():
    """ Get software info for inclusion into warcinfo """
    return {
            'platform': platform.platform (),
            'python': {
                'implementation': platform.python_implementation(),
                'version': platform.python_version (),
                'build': platform.python_build ()
                },
            'self': getRequirements (__package__)
        }

def getRequirements (dist):
    """ Get dependencies of a package.

    Figure out packages’ dependencies based on setup/distutils, then look at
    modules loaded and compute hashes of each loaded dependency.

    This does not and cannot protect against malicious people. It’s only
    purpose is to recreate this exact environment.
    """

    pending = {dist}
    have = set ()
    packages = []
    while pending:
        d = pkg_resources.get_distribution (pending.pop ())

        modules = list (filter (lambda x: x, d.get_metadata ('top_level.txt').split ('\n')))
        modhashes = {}
        # hash loaded modules
        for m in sys.modules.values ():
            f = getattr (m, '__file__', None)
            pkg = getattr (m, '__package__', None)
            # is loaded?
            if pkg in modules:
                if f and os.path.isfile (f):
                    with open (f, 'rb') as fd:
                        contents = fd.read ()
                        h = hashlib.new ('sha512')
                        h.update (contents)
                        modhashes[m.__name__] = {'sha512': h.hexdigest (), 'len': len (contents)}
                else:
                    modhashes[m.__name__] = {}

        # only if one of the packages’ modules is actually loaded
        if modhashes:
            packages.append ({'projectName': d.project_name, 'modules': modhashes, 'version': d.version})

        have.add (dist)
        pending.update (d.requires ())
        pending.difference_update (have)
    return packages

